/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cmath>

#include <iostream>
#include <fstream>

#include "s_solve.hpp"

#include "sGFX_solution.hpp"
#include "sGFX_problem.hpp"
#include "sGFX_pieces.hpp"

#include "sFOX_mainwin.hpp"
#include "sFOX_piecewin.hpp"
#include "sFOX_help.hpp"
#include "sFOX_icon.hpp"

//**************************************************************************

// Message Map for the Scribble App class
FXDEFMAP(SomaMainWin) SomaMainWinMap[]={
  
  //________Message_Type_________ID_____________________Message_Handler_______
  FXMAPFUNC(SEL_PAINT,            SomaMainWin::ID_CANVAS,    SomaMainWin::onExpose),
  FXMAPFUNC(SEL_CONFIGURE,        SomaMainWin::ID_CANVAS,    SomaMainWin::onConfigure),
  FXMAPFUNC(SEL_LEFTBUTTONPRESS,  SomaMainWin::ID_CANVAS,    SomaMainWin::onMouseClick),
  FXMAPFUNC(SEL_RIGHTBUTTONPRESS, SomaMainWin::ID_CANVAS,    SomaMainWin::onMouseClick),
  FXMAPFUNC(SEL_MIDDLEBUTTONPRESS,SomaMainWin::ID_CANVAS,    SomaMainWin::onMouseClick),
  FXMAPFUNC(SEL_MOTION,           SomaMainWin::ID_CANVAS,    SomaMainWin::onMouseMove),
  FXMAPFUNC(SEL_MOUSEWHEEL,       SomaMainWin::ID_CANVAS,    SomaMainWin::onMouseWheel),
  FXMAPFUNC(SEL_COMMAND,          SomaMainWin::ID_OPEN,      SomaMainWin::onCmdOpen), 
  FXMAPFUNC(SEL_COMMAND,          SomaMainWin::ID_SAVE,      SomaMainWin::onCmdSave),
  FXMAPFUNC(SEL_COMMAND,          SomaMainWin::ID_SOLVE,     SomaMainWin::onCmdSolve), 
  FXMAPFUNC(SEL_COMMAND,          SomaMainWin::ID_RESET,     SomaMainWin::onCmdReset),
  FXMAPFUNC(SEL_COMMAND,          SomaMainWin::ID_UNSOLVE,   SomaMainWin::onCmdUnsolve),
  FXMAPFUNC(SEL_COMMAND,          SomaMainWin::ID_HELP,      SomaMainWin::onCmdHelp),
  FXMAPFUNC(SEL_COMMAND,          SomaMainWin::ID_SHOWPIECES,SomaMainWin::onCmdShowPieces),
  FXMAPFUNC(SEL_COMMAND,          SomaMainWin::ID_PREV,      SomaMainWin::onCmdPrevNext),
  FXMAPFUNC(SEL_COMMAND,          SomaMainWin::ID_NEXT,      SomaMainWin::onCmdPrevNext)
  
};

//**************************************************************************

// Macro for the class hierarchy implementation
FXIMPLEMENT(SomaMainWin,FXMainWindow,SomaMainWinMap,ARRAYNUMBER(SomaMainWinMap))

//**************************************************************************    

SomaMainWin::SomaMainWin(FXApp* a):
             FXMainWindow(a,"Somatic",NULL,NULL,DECOR_ALL,0,0,800,600){
  
  FXIcon *openpzl=new FXGIFIcon(getApp(),fileopen); 
  FXIcon *savepzl=new FXGIFIcon(getApp(),filesaveas); 
  FXIcon *closepzl=new FXGIFIcon(getApp(),killobject); 
  FXIcon *solvepzl=new FXGIFIcon(getApp(),solvepuzzle); 
  FXIcon *resetpzl=new FXGIFIcon(getApp(),resetpuzzle);     
  FXIcon *unsolvepzl=new FXGIFIcon(getApp(),unsolvepuzzle); 
  FXIcon *helppzl=new FXGIFIcon(getApp(),helppuzzle);     
    
  statusbar = new FXStatusbar(this,LAYOUT_SIDE_BOTTOM|LAYOUT_FILL_X|FRAME_RAISED|STATUSBAR_WITH_DRAGCORNER); 
  
  FXHorizontalFrame *frame = new FXHorizontalFrame(this,LAYOUT_FILL_X|LAYOUT_FILL_Y|FRAME_RAISED,0,0,0,0,0,0,0,0);
   
  //LEFT pane to contain the glcanvas
  FXVerticalFrame *glcanvasFrame = new FXVerticalFrame(frame,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,5,1,5,5); 
  
  // RIGHT pane for the buttons
  FXVerticalFrame *buttonFrame=new FXVerticalFrame(frame,LAYOUT_FILL_Y,0,0,0,0,0,5,5,5);
  
  FXVerticalFrame *glpanel=new FXVerticalFrame(glcanvasFrame,FRAME_SUNKEN|FRAME_THICK|LAYOUT_FILL_X|LAYOUT_FILL_Y|LAYOUT_TOP|LAYOUT_LEFT,0,0,0,0,0,0,0,0);
  
  // A Visual to drag OpenGL
  glvisual=new FXGLVisual(getApp(),VISUAL_DOUBLEBUFFER);
  
  // Drawing glcanvas
  glcanvas=new FXGLCanvas(glpanel,glvisual,this,ID_CANVAS,LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0);

  // "Reset" button
  FXButton *bt = new FXButton(buttonFrame,NULL,resetpzl,this,ID_RESET,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X,0,0,0,0,10,10,5,5);
  bt->setHelpText("Reset figure."); 
   
  // "UnSolve" button
  bt = new FXButton(buttonFrame,NULL,unsolvepzl,this,ID_UNSOLVE,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X,0,0,0,0,10,10,5,5);
  bt->setHelpText("Delete solution.");
  
  FXMatrix* algparam=new FXMatrix(buttonFrame,2,FRAME_RAISED|FRAME_THICK|LAYOUT_FILL_X,0,0,0,0,10,10,5,5);
  
  FXLabel* create = new FXLabel(algparam,"Randomize:",NULL,LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,4,4,2,2);
  create->setHelpText("Randomize, to get different solutions on each call."); 
  create = new FXLabel(algparam,"Solutions:",NULL,LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,4,4,2,2);
  create->setHelpText("Number of solutions to create, if possible."); 
  
  rand = new FXCheckButton(algparam,"",NULL,0,JUSTIFY_NORMAL|ICON_AFTER_TEXT|LAYOUT_LEFT|LAYOUT_CENTER_Y,0,0,0,0,0,4,2,2);
  rand->setCheck(FALSE);
  rand->setHelpText("Randomize, to get different solutions on each call.");
  
  getsolutions = new FXSpinner(algparam,3,NULL,0,SPIN_CYCLIC|FRAME_SUNKEN|FRAME_THICK|LAYOUT_LEFT|LAYOUT_CENTER_Y,0,0,0,0,4,4,2,2);
  getsolutions->setRange(1,999);
  getsolutions->setValue(1);
  getsolutions->setHelpText("Number of solutions to create, if possible.");
  
  // "Solve" button
  bt = new FXButton(buttonFrame,NULL,solvepzl,this,ID_SOLVE,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X,0,0,0,0,10,10,5,5);
  bt->setHelpText("Find a solution.");
  
  FXHorizontalFrame *prevnext = new FXHorizontalFrame(buttonFrame,LAYOUT_FILL_X,0,0,0,0,0,0,0,0);
  
  prev = new FXButton(prevnext,"Previous",NULL,this,ID_PREV,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,10,10,5,5);
  prev->setHelpText("Previous solution.");

  next = new FXButton(prevnext,"  Next  ",NULL,this,ID_NEXT,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X|LAYOUT_FILL_Y,0,0,0,0,10,10,5,5);
  next->setHelpText("Next solution.");
    
  // "load..." button
  bt = new FXButton(buttonFrame,NULL,openpzl,this,ID_OPEN,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X,0,0,0,0,10,10,5,5);
  bt->setHelpText("Load puzzle.");
  
  // "Save..." button
  bt = new FXButton(buttonFrame,NULL,savepzl,this,ID_SAVE,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X,0,0,0,0,10,10,5,5);
  bt->setHelpText("Save puzzle.");
  
  bt = new FXButton(buttonFrame,"Pieces",NULL,this,ID_SHOWPIECES,FRAME_THICK|FRAME_RAISED|LAYOUT_FILL_X,0,0,0,0,10,10,5,5);
  bt->setHelpText("Show / hide pieces.");
  
  // bottom right pane to contain the exit and help button
  FXHorizontalFrame *exitFrame=new FXHorizontalFrame(buttonFrame,LAYOUT_FILL_Y|LAYOUT_FILL_X,0,0,0,0,0,0,0,0);
   
  // Help button
  bt = new FXButton(exitFrame,NULL,helppzl,this,ID_HELP,FRAME_THICK|FRAME_RAISED|LAYOUT_BOTTOM|LAYOUT_FILL_X,0,0,0,0,10,10,5,5);
  bt->setHelpText("About Somatic.");
  
  // Exit button
  bt = new FXButton(exitFrame,NULL,closepzl,getApp(),FXApp::ID_QUIT,FRAME_THICK|FRAME_RAISED|LAYOUT_BOTTOM|LAYOUT_FILL_X,0,0,0,0,10,10,5,5);
  bt->setHelpText("Exit.");

  new FXLabel(statusbar," ",NULL,LAYOUT_RIGHT|LAYOUT_CENTER_Y|JUSTIFY_RIGHT);
  FXVerticalFrame *statframe =  new FXVerticalFrame(statusbar,FRAME_SUNKEN|JUSTIFY_RIGHT|LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,4,4,2,2);
  statsolutions = new FXLabel(statframe,"",NULL,LAYOUT_CENTER_X|LAYOUT_CENTER_Y);
  statframe =  new FXVerticalFrame(statusbar,FRAME_SUNKEN|JUSTIFY_RIGHT|LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,4,4,2,2);
  statsolution = new FXLabel(statframe,"",NULL,LAYOUT_CENTER_X|LAYOUT_CENTER_Y);
  new FXLabel(statusbar,"Solutions:",NULL,LAYOUT_RIGHT|LAYOUT_CENTER_Y|JUSTIFY_RIGHT);
  statframe = new FXVerticalFrame(statusbar,FRAME_SUNKEN|JUSTIFY_RIGHT|LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,4,4,2,2);
  statmaxblocks = new FXLabel(statframe,"",NULL,LAYOUT_CENTER_X|LAYOUT_CENTER_Y);
  statframe = new FXVerticalFrame(statusbar,FRAME_SUNKEN|JUSTIFY_RIGHT|LAYOUT_RIGHT|LAYOUT_CENTER_Y,0,0,0,0,4,4,2,2);
  statblocks = new FXLabel(statframe,"",NULL,LAYOUT_CENTER_X|LAYOUT_CENTER_Y);
  new FXLabel(statusbar,"Cubes:",NULL,LAYOUT_RIGHT|LAYOUT_CENTER_Y|JUSTIFY_RIGHT);
  statframe = new FXVerticalFrame(statusbar,FRAME_SUNKEN|JUSTIFY_RIGHT|LAYOUT_RIGHT|LAYOUT_CENTER_Y|LAYOUT_FILL_Y,0,0,0,0,4,4,2,2);
  statpieces = new FXLabel(statframe,"",NULL,LAYOUT_CENTER_X|LAYOUT_CENTER_Y);
  new FXLabel(statusbar,"Pieces:",NULL,LAYOUT_RIGHT|LAYOUT_CENTER_Y|JUSTIFY_RIGHT);

  //read a start problem   
  std::ifstream istrm("../files/pieces/start.pzl");
  if(!istrm){
    std::cerr << "'../files/pieces/start.pzl' not found. Exiting." << std::endl;
    exit(-1);
  }
  
  if(!(p = SomaProb::create_problem(istrm))){ 
    std::cerr << "Could not read from '../files/pieces/start.pzl'. Exiting." 
              << std::endl;
    exit(-1);
  }
  
  // create gfx object for problem 
  if (p->is_solved()) p_gfx = new SomaSolGFX(p);
  else p_gfx = new SomaProbGFX(p);

  if(glcanvas->makeCurrent()){
    p_gfx->resize(glcanvas->getWidth(),glcanvas->getHeight());
    glcanvas->makeNonCurrent();
  }
  
  piecewin = 0;
  
  lastfile = "";
  lastdirectory = "../files";
  
  updateStatus();   
}

//**************************************************************************

SomaMainWin::~SomaMainWin(){
  delete p_gfx;
  delete p;
}

//**************************************************************************

// Create and initialize
void SomaMainWin::create(){
  FXMainWindow::create();
  show(PLACEMENT_SCREEN);
  
  // init gl stuff
  if(glcanvas->makeCurrent()){
    SomaBaseGFX::glInit();
    glcanvas->makeNonCurrent();
  }
}

//**************************************************************************

// Widget has been resized
long SomaMainWin::onConfigure(FXObject*,FXSelector,void*){
  if(glcanvas->makeCurrent()){
    p_gfx->resize(glcanvas->getWidth(),glcanvas->getHeight());
    glcanvas->makeNonCurrent();
  }
  
  return 1;
}

//**************************************************************************

// Widget needs repainting
long SomaMainWin::onExpose(FXObject*,FXSelector,void*){
  drawScene();
  return 1;
}
 
//**************************************************************************

long SomaMainWin::onMouseClick(FXObject *,FXSelector selector,void* ptr){
  FXEvent *e = (FXEvent*)ptr;
  
  SomaBaseGFX::MOUSEBUTTON button = SomaBaseGFX::NONE;

  switch(SELTYPE(selector)){
    case SEL_LEFTBUTTONPRESS:
      button = SomaBaseGFX::LEFT;
      break;
    case SEL_MIDDLEBUTTONPRESS:
      button = SomaBaseGFX::MIDDLE;
      break;
    case SEL_RIGHTBUTTONPRESS:
      button = SomaBaseGFX::RIGHT;
      break;
  }
  
  if(glcanvas->makeCurrent()){
    p_gfx->mouseClick(e->win_x,e->win_y,button);
    glcanvas->makeNonCurrent();
  }
  
  updateStatus();
  drawScene();
  return 1;
  
}

//**************************************************************************

long SomaMainWin::onMouseMove(FXObject*,FXSelector,void* ptr){
  FXEvent *e = (FXEvent*)ptr;

  SomaBaseGFX::MOUSEBUTTON button = SomaBaseGFX::NONE;
  
  
  if(e->state & RIGHTBUTTONMASK) button = SomaBaseGFX::RIGHT;
  else if(e->state & MIDDLEBUTTONMASK) button = SomaBaseGFX::MIDDLE;
  else if(e->state & LEFTBUTTONMASK) button = SomaBaseGFX::LEFT;
  
  if(button != SomaBaseGFX::NONE){
    
    if(glcanvas->makeCurrent()){
      p_gfx->mouseMove(e->win_x - e->last_x,e->win_y - e->last_y,button);
      glcanvas->makeNonCurrent();
    }
   
    drawScene();
  }
  
 
  
  return 1;
}

//**************************************************************************

long SomaMainWin::onMouseWheel(FXObject*,FXSelector,void* ptr){
  FXEvent *e = (FXEvent*) ptr;
  
  p_gfx->mouseMove(-e->code/12,-e->code/12,SomaBaseGFX::MIDDLE);

  drawScene();
  return 1;
}

//**************************************************************************

void SomaMainWin::updateStatus(const char *const str){

  if(p->is_solved()){
    unsigned int solutions = p->get_solution()->size();
    char *buffer = new char[(int) log10(solutions)+2];
    sprintf(buffer,"%u",solutions);
    statsolutions->setText(buffer);
    delete[] buffer;
    
    solutions = p->get_solution()->getActiveSolution()+1;
    buffer = new char[(int) log10(solutions+1)+2];
    sprintf(buffer,"%u",solutions);
    statsolution->setText(buffer);
    delete[] buffer;
  }else{
    statsolutions->setText(str);
    statsolution->setText("0");
  }
  
  const unsigned int nb = p->get_num_blocks();
  const unsigned int mb = p->get_max_blocks();

  statblocks->setText(FXStringFormat("%i",nb));
  statmaxblocks->setText(FXStringFormat("%i",mb));
  statpieces->setText(FXStringFormat("%i",p->get_num_pieces()));
}

//**************************************************************************

long SomaMainWin::onCmdReset(FXObject*,FXSelector,void*){
  p->unsolve();
  p->clear_blocks();
  
  SomaSingleGFX *swap = p_gfx;
  p_gfx = new SomaProbGFX(p,swap);
  delete swap;

  if(glcanvas->makeCurrent()){
    p_gfx->resize(glcanvas->getWidth(),glcanvas->getHeight());
    glcanvas->makeNonCurrent();
  }  
  
  if(piecewin && piecewin->shown()) piecewin->onExpose(NULL,0,NULL);
  drawScene();
  updateStatus();
  return 1;
}

//**************************************************************************

long SomaMainWin::onCmdOpen(FXObject*,FXSelector,void*){  
  FXFileDialog open(this,"Open");
  
  open.setPatternList("Somatic puzzles (*.pzl)\nAll Files (*)");
  open.setDirectory(lastdirectory.c_str());
  
  if(open.execute()){
    std::ifstream istrm(open.getFilename().text());
    
    if(!istrm || !p->read(istrm)){
      std::cerr << "Could not read file '" 
                << open.getFilename().text() << "'." << std::endl;
    }else{
      istrm.close();
      lastfile = open.getFilename().text();
      lastdirectory = open.getDirectory().text();
    }
     

    if (p->is_solved()) p_gfx = new SomaSolGFX(p);
    else p_gfx = new SomaProbGFX(p);
    
    if(piecewin){
      piecewin->hide();
      piecewin->destroy();
      delete piecewin;
      piecewin = 0;
    }

    if(glcanvas->makeCurrent()){
      p_gfx->resize(glcanvas->getWidth(),glcanvas->getHeight());
      glcanvas->makeNonCurrent();
    }
 
    drawScene();
    updateStatus();
  }
  
  return 1;
}

//**************************************************************************

long SomaMainWin::onCmdSave(FXObject*,FXSelector,void*){
  FXFileDialog save(this,"Save");
  
  save.setFilename(lastfile.c_str());
  save.setDirectory(lastdirectory.c_str());
  save.setPatternList("Somatic puzzles (*.pzl)\nAll Files (*)");
  
  if(save.execute()){
    std::ofstream ostrm(save.getFilename().text());
    
    if(!ostrm){
      std::cerr << "Could not open file '" << save.getFilename().text()
                << "' for writing. " << std::endl;
    }else{
     ostrm << *p;
     ostrm.close();
     lastfile = save.getFilename().text();
     lastdirectory = save.getDirectory().text();
    }
  }
  
  return 1;
}

//**************************************************************************

long SomaMainWin::onCmdHelp(FXObject*,FXSelector,void*){
  SomaHelp hd(this);
  hd.execute();
    
  return 1;
}

//**************************************************************************

long SomaMainWin::onCmdShowPieces(FXObject*,FXSelector,void*){
   
  if(!piecewin){
    piecewin = new SomaPieceWin(this,p,glcanvas);
    piecewin->create();
  }else{
    if(piecewin->shown()) piecewin->hide(); 
    else piecewin->show();
  }
  
  return 1;
}

//**************************************************************************

long SomaMainWin::onCmdSolve(FXObject*,FXSelector,void*){

  if (!p->is_solved()){
    
    SomaSolve *solver = SomaSolve::createSolver(p);
    if(solver == 0){
      updateStatus("Figure to big for solver !");
    }else{
      
      const int numsol = getsolutions->getValue();
      bool randsol = rand->getCheck();
      
      solver->solve(numsol,randsol);
      delete solver;
    
      if(p->is_solved()){

        SomaSingleGFX* swap = p_gfx;
        p_gfx = new SomaSolGFX(p,swap);
        delete swap;

        if(glcanvas->makeCurrent()){
          p_gfx->resize(glcanvas->getWidth(),glcanvas->getHeight());
          glcanvas->makeNonCurrent();
        }
        
        if(piecewin && piecewin->shown()) piecewin->onExpose(NULL,0,NULL);
        updateStatus();
        drawScene();      
      }else{
        updateStatus("Unsolvable !");
      }
    }
        
  }

  return 1;
}

//**************************************************************************

long SomaMainWin::onCmdPrevNext(FXObject*,FXSelector selector,void*){
  
  if(p->is_solved()){

    switch(SELID(selector)){
      case ID_PREV:
        p->get_solution()->prevSolution();
        break;
      case ID_NEXT:
        p->get_solution()->nextSolution();
        break;
    }
  }

  if(piecewin && piecewin->shown()) piecewin->onExpose(NULL,0,NULL);
  updateStatus();
  drawScene();
  return 1;
}
  
//**************************************************************************

long SomaMainWin::onCmdUnsolve(FXObject*,FXSelector,void*){
  if (p->is_solved()){

    p->unsolve();
   
    SomaSingleGFX* swap = p_gfx;
    p_gfx = new SomaProbGFX(p,swap); 
    delete swap;

    if(glcanvas->makeCurrent()){
      p_gfx->resize(glcanvas->getWidth(),glcanvas->getHeight());
      glcanvas->makeNonCurrent();
    }
    
    if(piecewin && piecewin->shown()) piecewin->onExpose(NULL,0,NULL);
    updateStatus();
    drawScene();
  }
 
  return 1;
}

//**************************************************************************

void SomaMainWin::drawScene(){
  if(glcanvas->makeCurrent()){
    p_gfx->draw();
    if(glvisual->isDoubleBuffer()) glcanvas->swapBuffers();
    glcanvas->makeNonCurrent(); 
  }
}

//**************************************************************************

