/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef soma_fox_mainwin
#define soma_fox_mainwin

#define HAVE_OPENGL

#include <fx.h>
#include <fx3d.h>

#include <string>

#include "s_problem.hpp"

#include "sGFX_single.hpp"
#include "sFOX_piecewin.hpp"

//**************************************************************************

// Event Handler Object
class FXAPI SomaMainWin : public FXMainWindow {
  FXDECLARE(SomaMainWin)
  
private:
  
  FXGLCanvas       *glcanvas;              // GL canvas to draw into
  FXGLVisual       *glvisual;              // OpenGL visual
  FXStatusbar      *statusbar;
  FXLabel          *statblocks,*statmaxblocks,*statsolution,*statsolutions,*statpieces; // place for number of actual blocks
  FXButton         *next,*prev;
  FXCheckButton    *rand;
  FXSpinner        *getsolutions;
  SomaProb         *p;                     // the problem object
                                           // graphical representation of it
  SomaSingleGFX    *p_gfx;

  SomaPieceWin    *piecewin;
  
  std::string lastfile,lastdirectory;  //last file opened  
  
  void drawScene();
  void updateStatus(const char *const str = "0"); // update status
  void create();                           // Initialize

protected:
  SomaMainWin(){}
  
public:

  // We define additional ID's, starting from the last one used by the base 
  // class+1. This way, we know the ID's are all unique for this particular 
  // target.
  enum{
    ID_CANVAS=FXMainWindow::ID_LAST,
    ID_OPEN,
    ID_SAVE,
    ID_RESET,
    ID_SOLVE,
    ID_UNSOLVE,
    ID_HELP,
    ID_SHOWPIECES,
    ID_LAST,
    ID_PREV,
    ID_NEXT
  };
  
  // Message handlers
  long onMouseClick(FXObject*,FXSelector,void*);
  long onMouseMove(FXObject*,FXSelector,void*);
  long onMouseWheel(FXObject*,FXSelector,void*);
  long onExpose(FXObject*,FXSelector,void*);
  long onConfigure(FXObject*,FXSelector,void*);
  long onCmdOpen(FXObject*,FXSelector,void*); 
  long onCmdSave(FXObject*,FXSelector,void*);
  long onCmdSolve(FXObject*,FXSelector,void*);
  long onCmdReset(FXObject*,FXSelector,void*); 
  long onCmdUnsolve(FXObject*,FXSelector,void*);
  long onCmdHelp(FXObject*,FXSelector,void*);
  long onCmdShowPieces(FXObject*,FXSelector,void*);
  long onCmdPrevNext(FXObject*,FXSelector,void*);
  
  //constructor
  SomaMainWin(FXApp* a);

  // destructor
  virtual ~SomaMainWin();
};

//**************************************************************************

#endif

