/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "sGFX_cube.hpp"

#define SOMA_CORNER  0.85
#define SOMA_TES 5

//**************************************************************************
  
unsigned int SomaCube::bodylst     = 0;
unsigned int SomaCube::borderlst   = 0;
unsigned int SomaCube::quadlst     = 0;
unsigned int SomaCube::edgelst     = 0;
unsigned int SomaCube::cornerlst   = 0;
unsigned int SomaCube::picklst     = 0;
unsigned int SomaCube::pickquadlst = 0;

//**************************************************************************
 
SomaPoint SomaCube::bisect(const SomaPoint& a,const SomaPoint& b){
  SomaPoint c = bisection(a,b);
  c.p.subtract(SOMA_CORNER,SOMA_CORNER,SOMA_CORNER);
  if(c.p.twoNorm() > 0) c.p /= c.p.twoNorm();
  c.n = c.p;
  c.p *= (1-SOMA_CORNER);
  c.p.add(SOMA_CORNER,SOMA_CORNER,SOMA_CORNER);

  return c;  
}

//**************************************************************************

void SomaCube::makeQuad(){
  const double trans = ((double) SOMA_TES)/2.0;
  const double scale = SOMA_CORNER/(trans*2.0);

  glNewList(quadlst,GL_COMPILE);
  
  glNormal3f(0,1,0);
  
  for (unsigned int i=0;i<SOMA_TES;i++){     
    glBegin(GL_QUAD_STRIP);
    for(unsigned int j=0;j<=SOMA_TES;j++){ 
      glVertex3d((i+1-trans)*scale, 0.5, (j-trans)*scale);
      glVertex3d((i-trans)*scale, 0.5, (j-trans)*scale);      
    }
    glEnd();
  }
  
  glEndList();
}

//**************************************************************************

void SomaCube::makeBody(){  
  glNewList(bodylst,GL_COMPILE);

  for(int j=0;j<6;j++){
    glPushMatrix();
    switch(j){
      case 1:
        glRotated( 90,1,0,0);
        break;
      case 2:
        glRotated(-90,1,0,0);
        break;
      case 3:
        glRotated(180,1,0,0);
        break;
      case 4:
        glRotated( 90,0,0,1);
        break;
      case 5:
        glRotated(-90,0,0,1);
    }
    
    glCallList(quadlst);
    glPopMatrix();
  }
  glEndList();
}

//**************************************************************************

void SomaCube::makeEdge(){
  SomaPoint a(SOMA_CORNER,SOMA_CORNER,1,          0,0,1);
  SomaPoint b(1,          SOMA_CORNER,SOMA_CORNER,1,0,0);
  
  SomaPoint c = bisect(a,b);
  
  a.p*= 0.5; b.p*=0.5; c.p*=0.5;

  const double trans = ((double) SOMA_TES)/2.0;
  const double scale = SOMA_CORNER/(trans*2.0);
  
  glNewList(edgelst,GL_COMPILE);
   
    glBegin(GL_QUAD_STRIP);
    for(unsigned int j=0;j<=SOMA_TES;j++){
      glNormal3d(a.n.x,a.n.y,a.n.z);
      glVertex3d(a.p.x,(j-trans)*scale,a.p.z);
      glNormal3d(c.n.x,c.n.y,c.n.z);
      glVertex3d(c.p.x,(j-trans)*scale,c.p.z);      
    }
    glEnd();

    glBegin(GL_QUAD_STRIP);
    for(unsigned int j=0;j<=SOMA_TES;j++){
      glNormal3d(c.n.x,c.n.y,c.n.z);
      glVertex3d(c.p.x,(j-trans)*scale,c.p.z);      
      glNormal3d(b.n.x,b.n.y,b.n.z);
      glVertex3d(b.p.x,(j-trans)*scale,b.p.z);
    }
    glEnd();

 
  glEndList();
}

//**************************************************************************

void SomaCube::makeCorner(){
  SomaPoint a(SOMA_CORNER,SOMA_CORNER,1,          0,0,1);
  SomaPoint b(1,          SOMA_CORNER,SOMA_CORNER,1,0,0);
  SomaPoint c(SOMA_CORNER,1,          SOMA_CORNER,0,1,0);
  
  SomaPoint d = bisect(a,b);
  SomaPoint e = bisect(b,c);
  SomaPoint f = bisect(c,a);
  
  a.p*= 0.5; b.p*=0.5; c.p*=0.5;
  d.p*= 0.5; e.p*=0.5; f.p*=0.5;
  
  glNewList(cornerlst,GL_COMPILE);
    
    glBegin(GL_TRIANGLE_STRIP);
      glNormal3d(b.n.x,b.n.y,b.n.z);
      glVertex3d(b.p.x,b.p.y,b.p.z);
      glNormal3d(e.n.x,e.n.y,e.n.z);
      glVertex3d(e.p.x,e.p.y,e.p.z);
      glNormal3d(d.n.x,d.n.y,d.n.z);
      glVertex3d(d.p.x,d.p.y,d.p.z);
      glNormal3d(f.n.x,f.n.y,f.n.z);
      glVertex3d(f.p.x,f.p.y,f.p.z);
      glNormal3d(a.n.x,a.n.y,a.n.z);
      glVertex3d(a.p.x,a.p.y,a.p.z);     
    glEnd();
  
    glBegin(GL_TRIANGLES);
      glNormal3d(f.n.x,f.n.y,f.n.z);
      glVertex3d(f.p.x,f.p.y,f.p.z);     
      glNormal3d(e.n.x,e.n.y,e.n.z);
      glVertex3d(e.p.x,e.p.y,e.p.z);
      glNormal3d(c.n.x,c.n.y,c.n.z);
      glVertex3d(c.p.x,c.p.y,c.p.z);
    glEnd();
 
  glEndList();
}

//**************************************************************************

void SomaCube::makeBorder(){
  glNewList(borderlst,GL_COMPILE);
  glPushMatrix();    
  for(int j=0;j<4;j++){
    glRotated(90,1,0,0);
    glPushMatrix();
    for(int k=0;k<3;k++){
      glCallList(edgelst);
      glRotated(-90,0,0,1);
    }
    glPopMatrix();
  }
  glPopMatrix();
  
  
  for(int j=0;j<8;j++){
    glPushMatrix();
    switch(j){
      case 1:
        glRotated( 90,1,0,0);
        break;
      case 2:
        glRotated(-90,1,0,0);
        break;
      case 3:
        glRotated(180,1,0,0);
        break;
      case 4:
        glRotated( 90,0,0,1);
        break;
      case 5:
        glRotated(180,0,0,1);
        break;
      case 6:
        glRotated(180,0,1,0);
        break;
      case 7:
        glRotated( 90,1,0,0);
        glRotated(180,0,0,1);
    }
    
    glCallList(cornerlst);
    glPopMatrix();
  }
 
  glEndList();
}

//**************************************************************************

void SomaCube::makePick(){
  glNewList(pickquadlst,GL_COMPILE);
    glBegin(GL_QUADS);
      glVertex3d(0.5,-0.5,-0.5);
      glVertex3d(0.5, 0.5,-0.5);
      glVertex3d(0.5, 0.5, 0.5);
      glVertex3d(0.5,-0.5, 0.5); 
    glEnd();
  glEndList();
  
  glNewList(picklst,GL_COMPILE);
   for(unsigned int j=0;j<6;j++){
    glPushMatrix();
    switch(j){
      case 0:
        glPushName(XP);
        break;
      case 1:
        glPushName(ZM);
        glRotated( 90,0,1,0);
        break;
      case 2:
        glPushName(ZP);
        glRotated(-90,0,1,0);
        break;
      case 3:
        glPushName(XM);
        glRotated(180,0,1,0);
        break;
      case 4:
        glPushName(YP);
        glRotated( 90,0,0,1);
        break;
      case 5:
        glPushName(YM);
        glRotated(-90,0,0,1);
    }
    
    glCallList(pickquadlst); 
    glPopName();
    glPopMatrix();
   }
  glEndList();  
}

//**************************************************************************

void SomaCube::makeAll(){
  static unsigned int inited = 0;
  if(inited) return;
  inited = 1;
  
  bodylst     = glGenLists(7);  
  borderlst   = bodylst+1;
  quadlst     = bodylst+2;
  edgelst     = bodylst+3;
  cornerlst   = bodylst+4;
  picklst     = bodylst+5;
  pickquadlst = bodylst+6;
  
  makeQuad();
  makeBody();
  makeEdge();
  makeCorner();
  makeBorder();
  makePick();

#ifdef SOMA_VERBOSE
  std::cout << "SomaCube display lists initialized." << std::endl;
#endif
}

//**************************************************************************

SomaCube::SomaCube(){
  pos = SomaVector<double>(0,0,0);
  color = SomaVector<double>(0.5,0.5,0.5);
}

//**************************************************************************

SomaCube::SomaCube(const SomaVector<double>& p,const SomaVector<double>& c):pos(p),color(c){}

//**************************************************************************

SomaCube::SomaCube(const SomaCube& other):pos(other.pos),color(other.color){}

//**************************************************************************

void SomaCube::draw(const GLenum mode)const{
  
  glPushMatrix();
  glTranslated(pos.x,pos.y,pos.z); 

  if(mode == GL_RENDER){   
    glColor3d(color.x,color.y,color.z);
    glCallList(bodylst);  
    //const SomaVector<double> c = 0.8*color;
    //glColor3d(c.x,c.y,c.z);
    glCallList(borderlst);      
  }else if(mode == GL_SELECT) {
    glCallList(picklst);        
  }
  glPopMatrix();
  
}

//**************************************************************************

