/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef soma_gfx_cube
#define soma_gfx_cube

#define HAVE_OPENGL

#include <fx.h>
#include <fx3d.h>

#include "s_vector.hpp"
#include "sGFX_pnt.hpp"

//**************************************************************************

class SomaCube{
private:
  static unsigned int bodylst;
  static unsigned int borderlst;
  static unsigned int quadlst;
  static unsigned int edgelst;
  static unsigned int cornerlst;
  static unsigned int pickquadlst;
  static unsigned int picklst;
  
  static SomaPoint bisect(const SomaPoint& a,const SomaPoint& b);
  
  static void makeQuad();
  static void makeBody();
  static void makeEdge();
  static void makeCorner();
  static void makeBorder();
  static void makePick();
  
  SomaVector<double> pos;
  SomaVector<double> color;
public:

  enum{XP=1,XM,YP,YM,ZP,ZM};
  
  SomaCube();
  SomaCube(const SomaVector<double>& p,const SomaVector<double>& c);
  SomaCube(const SomaCube& other);
  
  static void makeAll();
  
  inline ~SomaCube(){}
  
  void draw(const GLenum mode)const;

  inline void setPosition(const SomaVector<double>& p){pos = p;}
  inline SomaVector<double> getPosition()const{return pos;}
  inline void setColor(const SomaVector<double>& c){color = c;}
};

//**************************************************************************

#endif

