/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef soma_gfx_generic
#define soma_gfx_generic

#define HAVE_OPENGL

#include <fx.h>
#include <fx3d.h>

#include "sGFX_matrix.hpp"

//**************************************************************************

/** base class for an opengl object to be drawn on a glcanvas,
    all functions except construcotr and destructor require the
    openglcontext to be current */
class SomaBaseGFX{
protected:
  /** width and height of canvas */
  unsigned int width,height;

public:
  /** flags for which button was used */
  enum MOUSEBUTTON {LEFT,MIDDLE,RIGHT,NONE};
  
  /** set up lights and materials */
  static void glInit();
 
  /** destructor */
  virtual ~SomaBaseGFX(){}
  
  /** notify the object of a mouseclick */
  virtual void mouseClick(const unsigned int x,const unsigned int y,
                          const MOUSEBUTTON button)=0;
 
  /** notify the object of a mousemove */
  virtual void mouseMove(const int deltax,const int deltay,
                         const MOUSEBUTTON button)=0;
  
  /** notify the object that the underlying canvas size has changed */
  virtual void resize(const unsigned int width,const unsigned int height)=0;
  
  /** draw this object */
  virtual void draw()=0;
};

//**************************************************************************

#endif

