/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a rot of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cstring>
#include <cmath>

#include <iostream>

#include "sGFX_matrix.hpp"

//**************************************************************************

SomaMatrix::SomaMatrix(){ 
  setIdentity(); 
}

//**************************************************************************

SomaMatrix::SomaMatrix(const SomaMatrix& other){
  memcpy(this->mtr,other.mtr,16*sizeof(double));
}

//**************************************************************************

void SomaMatrix::rotate(const SomaMatrix::AXIS axis,const double deg){
  double rot[16] = {1,0,0,0,0,1,0,0,0,0,1,0,0,0,0,1};
  
  const double insin = sin(deg);
  const double incos = cos(deg);
  
  switch(axis){
    case X:
      rot[5] =  incos ;
      rot[6] =  insin ;
      rot[9] = -insin ;
      rot[10] = incos ;
      break;
    case Y:
      rot[0] =  incos ;
      rot[2] = -insin ;
      rot[8] =  insin ;
      rot[10] = incos ;
      break;
    case Z:
      rot[0] =  incos ;
      rot[1] =  insin ;
      rot[4] = -insin ;
      rot[5] =  incos ;
      break;
    default:
#ifdef SOMA_DEBUG
      std::cerr << "Error: Wrong Axis in SomaMatrix::rotate()." << std::endl;
#endif
      break;
  }
  
  const double res[16] = {
    rot[0]* mtr[0] + rot[4]*mtr[1]  + rot[8]* mtr[2]  + rot[12]* mtr[3],
    rot[1]* mtr[0] + rot[5]*mtr[1]  + rot[9]* mtr[2]  + rot[13]* mtr[3],
    rot[2]* mtr[0] + rot[6]*mtr[1]  + rot[10]*mtr[2]  + rot[14]* mtr[3],
    rot[3]* mtr[0] + rot[7]*mtr[1]  + rot[11]*mtr[2]  + rot[15]* mtr[3],
    rot[0]* mtr[4] + rot[4]*mtr[5]  + rot[8]* mtr[6]  + rot[12]* mtr[7],
    rot[1]* mtr[4] + rot[5]*mtr[5]  + rot[9]* mtr[6]  + rot[13]* mtr[7],
    rot[2]* mtr[4] + rot[6]*mtr[5]  + rot[10]*mtr[6]  + rot[14]* mtr[7],
    rot[3]* mtr[4] + rot[7]*mtr[5]  + rot[11]*mtr[6]  + rot[15]* mtr[7],
    rot[0]* mtr[8] + rot[4]*mtr[9]  + rot[8]* mtr[10] + rot[12]*mtr[11],
    rot[1]* mtr[8] + rot[5]*mtr[9]  + rot[9]* mtr[10] + rot[13]*mtr[11],
    rot[2]* mtr[8] + rot[6]*mtr[9]  + rot[10]*mtr[10] + rot[14]*mtr[11],
    rot[3]* mtr[8] + rot[7]*mtr[9]  + rot[11]*mtr[10] + rot[15]*mtr[11],
    rot[0]*mtr[12] + rot[4]*mtr[13] + rot[8]* mtr[14] + rot[12]*mtr[15],
    rot[1]*mtr[12] + rot[5]*mtr[13] + rot[9]* mtr[14] + rot[13]*mtr[15],
    rot[2]*mtr[12] + rot[6]*mtr[13] + rot[10]*mtr[14] + rot[14]*mtr[15],
    rot[3]*mtr[12] + rot[7]*mtr[13] + rot[11]*mtr[14] + rot[15]*mtr[15]
  };
  
  memcpy(mtr,res,16*sizeof(double));
}


//**************************************************************************

void SomaMatrix::setIdentity(){
  for (unsigned int i=0;i<16;i++){
    mtr[i] = 0;
  }
  mtr[0] = 1;
  mtr[5] = 1;
  mtr[10] = 1;
  mtr[15] = 1;
}

//**************************************************************************

