/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#ifndef soma_gfx_problem
#define soma_gfx_problem

#include "s_problem.hpp"

#include "sGFX_single.hpp"
#include "sGFX_cubearray.hpp"

//**************************************************************************

class SomaProbGFX:public SomaSingleGFX{
private:
  
  bool rotation_allowed;
 
  /** offset for centering graphic */
  SomaVector<double> offset;
  
  /** generate drawable cubes from problem object */
  void initCubes();
  
  /** add/remove a cube from pickingresult */
  void picked(const unsigned int name1, const unsigned int name2,
              const SomaBaseGFX::MOUSEBUTTON button);
  
  /** perform picking , return true if anything was hit */
  bool picking(const unsigned int x,const unsigned int y,
               const SomaBaseGFX::MOUSEBUTTON button);
  
  /** render object */
  void display(const GLenum mode);
public:
  
  /** constructor, create object from a problem */
  SomaProbGFX(SomaProb *const p);
  
  /** constructor, create object from a problem, keep rotation and distance 
      from other gfx object */
  SomaProbGFX(SomaProb *const p,const SomaSingleGFX *const gfx);
  
  // destructor
  ~SomaProbGFX();

  /** notify the object of a mouseclick, giving coordinates
      button and whether mouse key was moved up or down */
  void mouseClick(const unsigned int x,const unsigned int y,
                  const MOUSEBUTTON button);
 
  /** notify the object of a mousemove giving koordinate changes of mouse 
      pointer */
  void mouseMove(const int deltax,const int deltay,
                 const MOUSEBUTTON button); 
};

//**************************************************************************

#endif

