/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "sGFX_solution.hpp"
 
#include "s_piece.hpp"
#include "sGFX_colorchooser.hpp"
#include "sGFX_cube.hpp"

//**************************************************************************

SomaSolGFX::SomaSolGFX(SomaProb *const p,const SomaSingleGFX *const gfx):
                       SomaSingleGFX(p,gfx){
  scale = 1; 
  initCubes();
}

//**************************************************************************

SomaSolGFX::SomaSolGFX(SomaProb *const p):SomaSingleGFX(p){
  scale = 1;
  initCubes();
}

//**************************************************************************

SomaSolGFX::~SomaSolGFX(){}

//**************************************************************************

void SomaSolGFX::initCubes(){
  
  cubes.clear();
  
  SomaColorChooser rgb(pbl->get_num_pieces());

  SomaVector<double> foff(pbl->center());

  for(unsigned int tcnt=0;tcnt < pbl->get_num_pieces();tcnt++){

    
    const SomaPosOri *const posori = pbl->get_solution()->get()->get(tcnt);
    
    if(posori){
      
      SomaPiece piece(*pbl->get_piece(tcnt));
      SomaVector<double> ppos(posori->pos);
      piece.setOrientation(posori->ori);
      
      //get cube from this piece that has the largest distance to center
      //of figure
      SomaVector<double> poff(piece.getCube(0));
      for(unsigned int bcnt=1;bcnt< piece.getNumCubes();bcnt++){
        SomaVector<double> npoff(piece.getCube(bcnt));
        if((foff+ppos+poff).twoNorm() < (foff+ppos+npoff).twoNorm()){
          poff = npoff;
        }
      }

      for(unsigned int bcnt=0;bcnt< piece.getNumCubes();bcnt++){
    
        SomaVector<double> bpos(piece.getCube(bcnt));
        cubes.insert((foff+ppos+poff)*scale-poff+bpos,rgb.get(tcnt));
        
      } 
    }  
  }
  
  curr_solution = pbl->get_solution()->getActiveSolution();
}

//**************************************************************************

void SomaSolGFX::display(const GLenum mode){
  
  if(curr_solution != pbl->get_solution()->getActiveSolution()) initCubes();
  
  if (mode == GL_RENDER){
    for(unsigned int i=0;i < cubes.size();i++){
      cubes.get(i)->draw(mode);  
    }
  }
}

//**************************************************************************

void SomaSolGFX::mouseClick(const unsigned int x,const unsigned int y,
                            const MOUSEBUTTON button){

}

//**************************************************************************

void SomaSolGFX::mouseMove(const int deltax,const int deltay,
                           const MOUSEBUTTON button){
   if(button == MIDDLE){
    dist += 0.2 * deltay; 
    if(dist < zoomstart-zoomrange) dist = zoomstart-zoomrange;
    if(dist > zoomstart+zoomrange) dist = zoomstart+zoomrange; 
  }else if(button == LEFT){
    rotation.rotate(SomaMatrix::Y,0.015*deltax);
    rotation.rotate(SomaMatrix::X,0.015*deltay);
  }else if(button == RIGHT){
    scale += 0.01*deltax;
  
    if (scale < 1 ) scale = 1;
    if (scale > 3 ) scale = 3;

    initCubes();  
  }
}

//**************************************************************************

