/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "s_piece.hpp"

//**************************************************************************

const unsigned int lookup[24][6] = {{ 8,12,18,20, 3, 1},
                                    {11,13,17,21, 0, 2},
                                    {10,14,16,22, 1, 3},
                                    { 9,15,19,23, 2, 0},
                                    
                                    {12, 8,22,16, 7, 5},
                                    {15, 9,21,17, 4, 6},
                                    {14,10,20,18, 5, 7},
                                    {13,11,23,19, 6, 4},
  
                                    { 4, 0,11, 9,19,21},
                                    { 5, 3, 8,10,16,20},
                                    { 6, 2, 9,11,17,23},
                                    { 7, 1,10, 8,18,22},
                                    
                                    { 0, 4,15,13,23,17},
                                    { 1, 7,12,14,20,16},
                                    { 2, 6,13,15,21,19},
                                    { 3, 5,14,12,22,18},
                                    
                                    {17,19, 4, 2,13, 9},
                                    {18,16, 5, 1,12,10},
                                    {19,17, 6, 0,15,11},
                                    {16,18, 7, 3,14, 8},
                                    
                                    {21,23, 0, 6, 9,13},
                                    {22,20, 1, 5, 8,14},
                                    {23,21, 2, 4,11,15},
                                    {20,22, 3, 7,10,12}};

//**************************************************************************
                                  
const DIRECTION set_pos_lookup[24][24]={
{NO,YP,YM,YM,ZM,ZM,XP,ZM,ZM,ZM,ZM,ZM,ZP,ZP,ZM,ZP,ZM,ZP,XP,ZM,XM,ZM,ZM,ZP},
{YM,NO,YP,YM,ZM,XP,ZM,ZM,ZM,ZM,ZM,ZM,ZP,ZP,ZP,ZM,ZP,XP,ZM,ZM,ZP,XM,ZM,ZM},
{YM,YM,NO,YP,XP,ZM,ZM,ZM,ZM,ZM,ZM,ZM,ZM,ZP,ZP,ZP,XP,ZM,ZM,ZP,ZM,ZP,XM,ZM},
{YP,YM,YM,NO,ZM,ZM,ZM,XP,ZM,ZM,ZM,ZM,ZP,ZM,ZP,ZP,ZM,ZM,ZP,XP,ZM,ZM,ZP,XM},
{ZM,ZM,XP,ZM,NO,YP,YM,YM,ZP,ZP,ZM,ZP,ZM,ZM,ZM,ZM,XM,ZM,ZM,ZP,ZM,ZP,XP,ZM},
{ZM,XP,ZM,ZM,YM,NO,YP,YM,ZP,ZP,ZP,ZM,ZM,ZM,ZM,ZM,ZP,XM,ZM,ZM,ZP,XP,ZM,ZM},
{XP,ZM,ZM,ZM,YM,YM,NO,YP,ZM,ZP,ZP,ZP,ZM,ZM,ZM,ZM,ZM,ZP,XM,ZM,XP,ZM,ZM,ZP},
{ZM,ZM,ZM,XP,YP,YM,YM,NO,ZP,ZM,ZP,ZP,ZM,ZM,ZM,ZM,ZM,ZM,ZP,XM,ZM,ZM,ZP,XP},
{ZP,ZP,ZM,ZP,ZM,ZM,ZM,ZM,NO,XM,XP,XP,ZM,ZM,YM,ZM,ZM,ZM,ZP,YM,ZP,YP,ZM,ZM},
{ZP,ZM,ZP,ZP,ZM,ZM,ZM,ZM,XP,NO,XM,XP,ZM,YM,ZM,ZM,YM,ZM,ZM,ZP,YP,ZM,ZM,ZP},
{ZM,ZP,ZP,ZP,ZM,ZM,ZM,ZM,XP,XP,NO,XM,YM,ZM,ZM,ZM,ZP,YM,ZM,ZM,ZM,ZM,ZP,YP},
{ZP,ZP,ZP,ZM,ZM,ZM,ZM,ZM,XM,XP,XP,NO,ZM,ZM,ZM,YM,ZM,ZP,YM,ZM,ZM,ZP,YP,ZM},
{ZM,ZM,ZM,ZM,ZP,ZP,ZM,ZP,ZM,ZM,YM,ZM,NO,XM,XP,XP,ZP,YP,ZM,ZM,ZM,ZM,ZP,YM},
{ZM,ZM,ZM,ZM,ZP,ZM,ZP,ZP,ZM,YM,ZM,ZM,XP,NO,XM,XP,YP,ZM,ZM,ZP,YM,ZM,ZM,ZP},
{ZM,ZM,ZM,ZM,ZM,ZP,ZP,ZP,YM,ZM,ZM,ZM,XP,XP,NO,XM,ZM,ZM,ZP,YP,ZP,YM,ZM,ZM},
{ZM,ZM,ZM,ZM,ZP,ZP,ZP,ZM,ZM,ZM,ZM,YM,XM,XP,XP,NO,ZM,ZP,YP,ZM,ZM,ZP,YM,ZM},
{ZM,ZM,XM,ZP,XP,ZM,ZM,ZP,ZP,YP,ZM,ZM,ZM,YM,ZP,ZM,NO,ZM,ZM,ZP,YM,ZM,XP,ZM},
{ZM,XM,ZP,ZM,ZP,XP,ZM,ZM,ZM,ZP,YP,ZM,YM,ZP,ZM,ZM,ZP,NO,ZM,ZM,ZM,XP,ZM,YM},
{XM,ZP,ZM,ZM,ZM,ZP,XP,ZM,ZM,ZM,ZP,YP,ZP,ZM,ZM,YM,ZM,ZP,NO,ZM,XP,ZM,YM,ZM},
{ZP,ZM,ZM,XM,ZM,ZM,ZP,XP,YP,ZM,ZM,ZP,ZM,ZM,YM,ZP,ZM,ZM,ZP,NO,ZM,YM,ZM,XP},
{XP,ZM,ZM,ZP,ZM,ZM,XM,ZP,ZM,YM,ZP,ZM,ZP,YP,ZM,ZM,YM,ZM,XP,ZM,NO,ZM,ZM,ZP},
{ZP,XP,ZM,ZM,ZM,XM,ZP,ZM,YM,ZP,ZM,ZM,ZM,ZP,YP,ZM,ZM,XP,ZM,YM,ZP,NO,ZM,ZM},
{ZM,ZP,XP,ZM,XM,ZP,ZM,ZM,ZP,ZM,ZM,YM,ZM,ZM,ZP,YP,XP,ZM,YM,ZM,ZM,ZP,NO,ZM},
{ZM,ZM,ZP,XP,ZP,ZM,ZM,XM,ZM,ZM,YM,ZP,YP,ZM,ZM,ZP,ZM,YM,ZM,XP,ZM,ZM,ZP,NO}
};

//**************************************************************************

template <class T>
SomaCubeSet<T>::SomaCubeSet(){
  curr_or = 0;
  blocks = new std::vector<SomaVector<T> >(1);
  (*blocks)[0] = SomaVector<T>(0,0,0); 
}

//**************************************************************************

template <class T>
SomaCubeSet<T>::SomaCubeSet(const unsigned int nobin){
#ifdef SOMA_DEBUG
  if(nobin == 0){
    std::cerr << "Attempt to create piece with no "
              << "blocks in SomaCubeSet::SomaCubeSet()." 
              << std::endl;
  }
#endif
  curr_or = 0;
  blocks = new std::vector<SomaVector<T> >(nobin);
  
  for(unsigned int i=0;i<blocks->size();i++){
    (*blocks)[i] = SomaVector<T>(i,0,0);
  }
}

//**************************************************************************

template <class T>
SomaCubeSet<T>::SomaCubeSet(const SomaCubeSet<T>& other){
  curr_or = other.curr_or;
  blocks = new std::vector<SomaVector<T> >(*other.blocks);
}

//**************************************************************************

template <class T>
SomaCubeSet<T>::~SomaCubeSet(){
  delete blocks;
}

//**************************************************************************

template <class T>
void SomaCubeSet<T>::set(const SomaCubeSet<T>& other){

  if(this == &other) return;

  delete blocks;  
  curr_or = other.curr_or;
  blocks = new std::vector<SomaVector<T> >(*other.blocks);
}

//**************************************************************************

template <class T>
void SomaCubeSet<T>::setCube(const unsigned int block,
                                    const SomaVector<T>& xyz){
#ifdef SOMA_DEBUG
  if(block >= blocks->size()){
    std::cerr << "Array bounds violated in SomaCubeSet::setCube()." 
              << std::endl;
    return;
  }
#endif
  
  (*blocks)[block] = xyz;
}
  
//************************************************************************** 

template <class T>
SomaVector<T> SomaCubeSet<T>::getCube(const unsigned int block)const{
#ifdef SOMA_DEBUG
  if(block >= blocks->size()){
    std::cerr << "Array bounds violated in SomaCubeSet::getCube()." 
              << std::endl;
    return SomaVector<T>(0,0,0);
  }
#endif

  return (*blocks)[block];
}
  
//************************************************************************** 

template <class T>
void SomaCubeSet<T>::setOrientation(const unsigned int orientation){
#ifdef SOMA_DEBUG
  if(orientation >= ORIENTATIONS){
    std::cerr << "Wrong orientation in SomaCubeSet::setOrientation()." 
              << std::endl;
    return;
  }
#endif
  
  while(curr_or != orientation)  
    rotate(set_pos_lookup[curr_or][orientation]);
}

//**************************************************************************

template <class T>  
void SomaCubeSet<T>::rotate(const DIRECTION direction){
  T swt;
  
  switch (direction){
    case ZM:
    for(unsigned int i=0;i<blocks->size();i++){
      swt = (*blocks)[i].x;
      (*blocks)[i].x = (*blocks)[i].y;
      (*blocks)[i].y = -swt;
    }
    break;
    case ZP:
    for(unsigned int i=0;i<blocks->size();i++){
      swt = (*blocks)[i].y;
      (*blocks)[i].y = (*blocks)[i].x;
      (*blocks)[i].x= -swt;
    }   
    break;
    case XP:
    for(unsigned int i=0;i<blocks->size();i++){
      swt = (*blocks)[i].z;
      (*blocks)[i].z = (*blocks)[i].y;
      (*blocks)[i].y = -swt;
    }
    break;
    case XM:
    for(unsigned int i=0;i<blocks->size();i++){
      swt = (*blocks)[i].y;
      (*blocks)[i].y = (*blocks)[i].z;
      (*blocks)[i].z = -swt;
    }
    break;
    case YP:
    for(unsigned int i=0;i<blocks->size();i++){
      swt = (*blocks)[i].x;
      (*blocks)[i].x = (*blocks)[i].z;
      (*blocks)[i].z = -swt;
    }   
    break;
    case YM:
    for(unsigned int i=0;i<blocks->size();i++){
      swt = (*blocks)[i].z;
      (*blocks)[i].z = (*blocks)[i].x;
      (*blocks)[i].x = -swt;
    }
    break;
    default:
      std::cerr << "Unknown direction in SomaCubeSet::rotate()."
                << std::endl;
      return;
  }
  
  curr_or = lookup[curr_or][direction];
}

//**************************************************************************

template class SomaCubeSet<int>;
template class SomaCubeSet<double>;

//**************************************************************************
//**************************************************************************
//**************************************************************************

bool SomaPiece::zeroCube(const std::vector<SomaVector<int> > *const cubes){
  const SomaVector<int> zero(0,0,0);
  
  for(unsigned int i=0;i<cubes->size();i++){
    if((*cubes)[i] == zero) return true;
  }
  
  return false;
}

//**************************************************************************

bool SomaPiece::noDoubleCubes(const std::vector<SomaVector<int> > *const cubes){
  for(unsigned int i=0;i<cubes->size();i++){
    for(unsigned int j=i+1;j<cubes->size();j++){
      if((*cubes)[i] == (*cubes)[j]) return false;
    }
  }
  
  return true;
}

//**************************************************************************
  
bool SomaPiece::connected(const std::vector<SomaVector<int> > *const cubes){
  std::vector<bool> found(cubes->size());
  std::vector<SomaVector<int> > queue(cubes->size());
  SomaVector<int> xp,xm,yp,ym,zp,zm;
  unsigned int qs = 0;
  unsigned int qe = 0;
  
  for (unsigned int i=0;i<cubes->size();i++)
    found[i] = false;
    
    queue[qs] = (*cubes)[0];
    found[0] = true;
    qs++;
    
    while(qe != qs){    
      // get from queue
      xp = queue[qe];
      xp.add(1,0,0);
      xm = queue[qe];
      xm.subtract(1,0,0);
      
      yp = queue[qe];
      yp.add(0,1,0);
      ym = queue[qe];
      ym.subtract(0,1,0);
      
      zp = queue[qe];
      zp.add(0,0,1);
      zm = queue[qe];
      zm.subtract(0,0,1);
      
      qe++;
      
      //check six directions
      for(unsigned int i=0;i<cubes->size();i++){
        if(!found[i]){
          
          if(xp == (*cubes)[i]){
            queue[qs] = (*cubes)[i];
            found[i] = true;
            qs++;
          }else if(xm == (*cubes)[i]){
            queue[qs] = (*cubes)[i];
            found[i] = true;
            qs++;
          }else if(yp == (*cubes)[i]){
            queue[qs] = (*cubes)[i];
            found[i] = true;
            qs++;
          }else if(ym == (*cubes)[i]){
            queue[qs] = (*cubes)[i];
            found[i] = true;
            qs++;
          }else if(zp == (*cubes)[i]){
            queue[qs] = (*cubes)[i];
            found[i] = true;
            qs++;
          }else if(zm == (*cubes)[i]){
            queue[qs] = (*cubes)[i];
            found[i] = true;
            qs++;  
          }
          
        }
      }
    }
    
    for(unsigned int i=0;i<cubes->size();i++){
      if(!found[i]) return false;
    }
    return true;
}

//**************************************************************************

SomaVector<double> SomaPiece::getOffset()const{
  SomaVector<int> min((*blocks)[0]);
  SomaVector<int> max((*blocks)[0]);
  
  for(unsigned int i=1;i<blocks->size();i++){
    if(max.x < (*blocks)[i].x) max.x = (*blocks)[i].x;
    if(max.y < (*blocks)[i].y) max.y = (*blocks)[i].y;
    if(max.z < (*blocks)[i].z) max.z = (*blocks)[i].z;

    if(min.x > (*blocks)[i].x) min.x = (*blocks)[i].x;
    if(min.y > (*blocks)[i].y) min.y = (*blocks)[i].y;
    if(min.z > (*blocks)[i].z) min.z = (*blocks)[i].z;    
  }
  
  SomaVector<double> retval(min+max);
  retval /= 2;
  retval.negate();
  
  return retval;
}
  
//**************************************************************************

bool SomaPiece::valid()const{
  if(!zeroCube(blocks)) return false;
  if(!noDoubleCubes(blocks)) return false;
  if(!connected(blocks)) return false;
  return true;
}

//**************************************************************************

void SomaPiece::write(std::ostream& strm)const{
  strm << " { (";
  strm.width(3);
  strm << blocks->size() << ") ";
  for (unsigned int i=0;i<blocks->size();i++){
    strm << (*blocks)[i];
  }
  strm << " } " << std::endl;
}

//**************************************************************************

bool SomaPiece::read(std::istream& strm){
  char S;
  bool error = false;
  unsigned int newnob=0;
  std::vector<SomaVector<int> >* newblocks=0;
  
  S=' '; 
  strm >> S;
  
  if(S!='{'){
    std::cerr << "Error in SomaPiece::read(). '{' expected." 
              << std::endl;
    error = true;
    goto error_stop;
  }
  
  
  S=' '; 
  strm >> S;
  
  if(S=='('){
    
    strm >> newnob;
    
    if(newnob>0){
      newblocks = new std::vector<SomaVector<int> >(newnob);
    }else{
      std::cerr << "Number of blocks 0 in SomaPiece::read()." << std::endl;
      error = true;
      goto error_stop;
    }
  }else{
    std::cerr << "Error in SomaPiece::read(). '(' expected." 
              << std::endl;
    error = true;
    goto error_stop;
  }
  
  S=' '; 
  strm >> S;
  
  if(S!=')'){
    std::cerr << "Error in SomaPiece::read(). ')' expected." 
              << std::endl;
    error = true;
    goto error_stop;
  }
  
  
  for(unsigned int i=0;i<newblocks->size();i++){
    if(!(*newblocks)[i].read(strm)){
      error = true;
      goto error_stop;
    }
  }
  
  S=' '; 
  strm >> S;
  
  if(S!='}'){
    std::cerr << "Error in SomaPiece::read(). '}' expected." 
              << std::endl;
    error = true;
    goto error_stop;
  }

  // ( 0 , 0 , 0 ) block required
  if(!zeroCube(newblocks)){
    std::cerr << "Input sematically incorrect in SomaPiece::read()." 
              << std::endl << "( 0 , 0 , 0 ) block required." << std::endl;
    error = true;
    goto error_stop;
  }
      
  // no double blocks
  if(!noDoubleCubes(newblocks)){
    std::cerr << "Input semantically incorrect in SomaPiece::read()." 
              << std::endl << "No double blocks allowed." << std::endl;
    error = true;
    goto error_stop;
  }

  // check blocks for connectedness !
  if(!connected(newblocks)){ 
    std::cerr << "Input semantically incorrect in SomaPiece::read()." 
              << std::endl << "Blocks in one piece have to be connected."
              << std::endl;
    error = true;
    goto error_stop;
  }
    
  error_stop:
  
  if(error){
    
    if(newblocks) delete newblocks;
    return false;
    
  }else{
    
    delete blocks;
    blocks = newblocks;
    curr_or = 0;
    
    return true;
  }
}

//**************************************************************************

