/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
 
#ifndef soma_solve
#define soma_solve

#include "s_vector.hpp"
#include "s_solution.hpp"
#include "s_problem.hpp"
#include "s_knapsack.hpp"

#include <map>
#include <list>

//**************************************************************************

/// bitvector needs to be an unsigned integer datatype
typedef unsigned long long bitvector;

//**************************************************************************

class SomaSolve{
private:
  static bitvector flags[];
  static void initFlags();
  static void printT(const bitvector val);
  static unsigned int calc_blockmax(const SomaProb *const p);
  static SomaVector<int> calc_min(const SomaProb *const p);
  static SomaVector<int> calc_max(const SomaProb *const p,
                                  const SomaVector<int>& min,
                                  const unsigned int blockmax);

  SomaProb *const p;
  const unsigned int num_blocks;
  const unsigned int num_pieces;
  
  const unsigned int blockmax;
  const SomaVector<int> min,max;
  const unsigned int xmult,ymult;

  unsigned int* space;
  
  std::map<bitvector,SomaPosOri>* insmap;
  bitvector** insarr; // insertarray
  unsigned int* ial; // length of insertarrays
  unsigned int* iapn; // piecenumbers of insertarrays in problem coordinates
  
  unsigned int num_selected;
  bitvector** insarr2;
  unsigned int* ial2;
  unsigned int* iapn2;
  
  bitvector* inuse;
  bitvector inserted,solved;

  unsigned int num_solutions;
  unsigned int ta;
  const unsigned int* possible;
  SomaKnapsack* K;
  
  // connectCheck stuff
  const unsigned int val2,val3,val4,val5;
  unsigned int* queue;
  unsigned int* confound;
  unsigned int* knapsack_solvable;
  int* reverse_lookup;
  
  //coverCheck stuff
  bitvector* cover;

  //algorithm parameters
  unsigned int stop_solutions;
  unsigned int just_count;
    
  SomaCubeSet<double> cubes;
  unsigned int* permutation;
  
  unsigned int getPermutation(const SomaCubeSet<double>& source,
                              const SomaCubeSet<double>& target,
                              unsigned int *const permutation);
  void shortenList(std::list<bitvector>& inslist,
                   const unsigned int *const permutation);
                   
  void createLists();
  void initConnectCheck();
  unsigned int connectCheck();
  unsigned int coverCheck();
  void writeSolution();
  
  /** Constructor */
  SomaSolve(SomaProb *const p_in);  
  
  void recurse();
  void recurseSymmetryCheck();
  
public:
  /** create an instance of the class
      returns null if this solve class can not handle the given problem */
  static SomaSolve* createSolver(SomaProb *const p); 
  
  /** Destructor */
  ~SomaSolve();
  
  /** Run the algorithm. Stop after num_solutions solutions found.
      Giving a parameter of 0 will stop after all solutions. Setting randomize
      to 1 will distort internal tables some, so different solutions will be
      found first. */
  unsigned int solve(const unsigned int stop_solutions=1,
                     const unsigned int randomize=0,
                     const unsigned int just_count=0);
};

//**************************************************************************

#endif
