/*  Somatic - A soma puzzle solver.
 *  Copyright (C) 2003 Marc Moerig
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <cstdlib>
 
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>

#include "../source/s_problem.hpp"
#include "../source/s_solve.hpp"

//**************************************************************************

std::string inputfile;
std::string outputfile;
unsigned int randomization = 0;
unsigned int limit = UINT_MAX;
unsigned int count = 1;

//**************************************************************************

void printHelp(){
  
 std::cerr << std::endl
           << " usage:" << std::endl
           << " txtsomatic [options] [input file] [output file]" << std::endl
           << std::endl
           << " solver mode:" << std::endl
           << " --mode=[mode] | default --mode=count | short -m=[mode]" << std::endl
           << " [mode] can be 'count','c' or 'solve','s'." << std::endl
           << std::endl
           << " solution limit:" << std::endl
           << " --limit=[number] | default --limit=0 | short -l=[number]" << std::endl
           << " [number] can be any unsigned integer." << std::endl
           << std::endl
           << " randomization:" << std::endl
           << " --randomize=[boolean] | default --randomize=false | short -r=[boolean]" << std::endl
           << " [boolean] can be either 'true','1' or 'false','0'." << std::endl
           << std::endl
           << " For more information please read the README file." << std::endl
           << std::endl;
  exit(1);
}

//**************************************************************************

void recognize(std::string *const parameter){

        if(*parameter == "--randomize=true" || 
           *parameter == "--randomize=1" ||
           *parameter == "-r=true" ||
           *parameter == "-r=1"){
             
    *parameter = "-r=1";
    
  }else if(*parameter == "--randomize=false" ||
           *parameter == "--randomize=0" ||
           *parameter == "-r=false" ||
           *parameter == "-r=0"){
             
    *parameter = "-r=0";
    
  }else if(*parameter == "--mode=count" ||
           *parameter == "--mode=c" ||
           *parameter == "-m=count" ||
           *parameter == "-m=c"){
             
    *parameter = "-m=c";
    
  }else if(*parameter == "--mode=solve" ||
           *parameter == "--mode=s" ||
           *parameter == "-m=solve" ||
           *parameter == "-m=s"){
    
    *parameter = "-m=s";
    
  }else if(parameter->find("--limit=") == 0){
    parameter->erase(0,8);
  }else if(parameter->find("-l=") == 0){
    parameter->erase(0,3);
  }else{
    std::cerr << "Unknown parameter '"<< *parameter  << "'." << std::endl;
    printHelp();
  }
}

//**************************************************************************

void parseCommandLine(int argc, char *argv[]){
  
  if(argc == 0) printHelp();
  
  std::string *parameters = new std::string[argc];
  for(int i=0;i<argc;i++){
    parameters[i] = argv[i];
  }
  
  int i=0;
  for(;i<argc;i++){
    if(parameters[i][0] != '-') break;
  }
  
  if(argc-i == 1){
    inputfile = outputfile = parameters[i];
  }else if(argc-i == 2){
    inputfile = parameters[i];
    if(parameters[i+1][0] == '-'){
      std::string copy = parameters[i+1];
      recognize(parameters+i+1);
      std::cerr << "Parameter '" << copy << "' at wrong position." << std::endl;
      printHelp();
    }else outputfile = parameters[i+1];
  }else{
    std::cerr << "Input file missing." << std::endl;
    printHelp();
  }
  
  for(int j=0;j<i;j++){
    std::string copy = parameters[j];
    recognize(parameters+j);
    
    if(parameters[j] == "-r=1") randomization=1;
    else if(parameters[j] == "-r=0") randomization=0;
    else if(parameters[j] == "-m=c") count = 1;
    else if(parameters[j] == "-m=s") count = 0;
    else{
      
      std::stringstream strm(parameters[j]);
      strm >> limit;
      
      if(strm.fail() || ! strm.eof()){
        std::cerr << "'" << parameters[j] << "' is not a valid number in '" 
                  << copy << "'." << std::endl;
        printHelp();
      }
      
      if(limit == 0) limit = UINT_MAX;
      
    }
  }
    
  delete[] parameters;
}

//**************************************************************************

int main(int argc, char *argv[]){
  
  parseCommandLine(argc-1,argv+1);
  
  std::ifstream istrm(inputfile.c_str());
  if(!istrm){
    std::cout << "Could not open '" << inputfile << "' for reading." 
              << std::endl;
    exit(1);
  }
  
  SomaProb* prb = SomaProb::create_problem(istrm);
  istrm.close();
  
  if(!prb) exit(1);
  
  SomaSolve *solver = SomaSolve::createSolver(prb);
  if(!solver){
    std::cerr << "Problem to big, solver not startet." << std::endl;
  }else{
    std::cout << solver->solve(limit,randomization,count) 
              << " solution(s) found." << std::endl;
    delete solver;
  }
    
  if(!count){
    std::ofstream ostrm(outputfile.c_str());
    if(!ostrm){
      std::cerr << "Could not open '" << outputfile << "' for writing." 
                << std::endl;
      std::cout << *prb;
    }else{
      ostrm << *prb;
      ostrm.close();
    }
  }
  
  delete prb;
  
  return 0;
}

//**************************************************************************
